/*
 * Multicast Fastforward from multicast source to destination ports
 *
 * Copyright (C) 2006 AVM GmbH
 *
 */

#include <linux/config.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/spinlock.h>
#include <linux/slab.h>
#include <linux/if.h>
#include <linux/if_ether.h>
#include <linux/if_vlan.h>
#include <linux/if_pppox.h>
#include <linux/module.h>
#include <linux/vermagic.h>
#include <linux/compiler.h>
#include <linux/mcfastforward.h>

#ifndef LOCALAREA_MCAST
#define LOCALAREA_MCAST(x)  (((x) & htonl(0xFFFF0000)) == htonl(0xEFFF0000))
#endif

#define MAX_NETDRIVER	4

static struct mcfw_netdriver *mcfw_drivers[MAX_NETDRIVER];

static struct mcfw_statistic mcfw_stat;

/* --------------------------------------------------------------------- */

int mcfw_netdriver_register(struct mcfw_netdriver *drv)
{
   int i;
   for (i = 0; i < MAX_NETDRIVER; i++) {
	  if (drv == mcfw_drivers[i]) 
		 return -EEXIST;
   }
   for (i = 0; i < MAX_NETDRIVER; i++) {
	  if (mcfw_drivers[i] == 0) {
		 mcfw_drivers[i] = drv;
		 drv->index = i;
		 return 0;
	  }
   }
   return -ENFILE;
}

void mcfw_netdriver_unregister(struct mcfw_netdriver *drv)
{
   int i;
   for (i = 0; i < MAX_NETDRIVER; i++) {
	  if (drv == mcfw_drivers[i]) {
		 mcfw_drivers[i] = 0;
		 return;
	  }
   }
}

/* --------------------------------------------------------------------- */

struct mcsubcription;

struct mcquery {
   struct mcquery       *next;
   struct mcsubcription *sub;
   mcfw_portset          portset;
   struct timer_list     timer;
};

struct mcgroupmember {
   struct mcgroupmember  *next;
   struct mcfw_netdriver *drv;
   u32                    ipaddr;
   int                    port;
   int                    active;
};

struct mcgroup {
   struct mcgroup       *link; /* hashtab */
   u32                   group;
   u8                    groupmac[ETH_ALEN];
   struct mcsubcription {
      mcfw_portset           portset;
      u32                    group;
      struct mcquery        *queries;
   }                     subscriptions[MAX_NETDRIVER];
   struct mcgroupmember *members;
};

/* --------------------------------------------------------------------- */

#if 1
static char *portset2str(mcfw_portset *p, char *buf, size_t size)
{
   char *s = buf;
   size_t len = 0;
   int i;

   for (i=0; i < MCFW_MAX_PORT; i++) {
      if (mcfw_portset_port_is_in_set(p, i)) {
		 snprintf(s, size-len, "%s%d", s != buf ? "," : "", i);
		 s += strlen(s);
	  }
   }
   *s = 0;
   return buf;
}
#endif

/* --------------------------------------------------------------------- */

static spinlock_t group_lock = SPIN_LOCK_UNLOCKED;
#define MCGROUP_HASHSIZ 256
static struct mcgroup *grouphashtab[MCGROUP_HASHSIZ];
static kmem_cache_t *mcquery_cachep;

static void mcgroup_check_del(u32 group);

static void mcquery_timer_handler(unsigned long arg)
{
   struct mcquery *q = (struct mcquery *)arg;
   struct mcsubcription *sub;
   struct mcquery **pp;
   unsigned long flags;

   spin_lock_irqsave(&group_lock, flags);
   sub = q->sub;
   for (pp = &sub->queries; *pp && *pp != q; pp = &(*pp)->next)
	  ;
   if (*pp == q) {
	  *pp = q->next;
      mcfw_portset_portset_del(&sub->portset, q->portset);
	  kmem_cache_free(mcquery_cachep, q);
	  if (mcfw_portset_empty(&sub->portset))
	     mcgroup_check_del(sub->group);
   }
   spin_unlock_irqrestore(&group_lock, flags);
}

static int mcquery_add(struct mcsubcription *sub,
                       mcfw_portset portset, unsigned long maxdelay)
{
   struct mcquery **pp;
   struct mcquery *p;

   p = kmem_cache_alloc(mcquery_cachep, SLAB_ATOMIC);
   if (p == 0)
	  return -ENOMEM;
   p->sub = sub;
   p->portset = portset;
   p->next = 0;
   for (pp = &sub->queries; *pp; pp = &(*pp)->next)
	  ;
   *pp = p;
   mod_timer(&p->timer, jiffies + maxdelay);
   return 0;
}

static void mcquery_ctor(void *mem, kmem_cache_t *cache, unsigned long flags)
{
   struct mcquery *p = (struct mcquery *)mem;
   memset(p, 0, sizeof(struct mcquery));
   init_timer(&p->timer);
   p->timer.function = mcquery_timer_handler;
   p->timer.data	 = (unsigned long)p;
}

static int mcquery_init(void)
{
	mcquery_cachep = kmem_cache_create("mcquery",
					                   sizeof(struct mcquery),
					                   0, SLAB_HWCACHE_ALIGN,
					                   mcquery_ctor, NULL);
    return mcquery_cachep ? 0 : -ENOMEM;
}

static void mcquery_exit(void)
{
    (void)kmem_cache_destroy(mcquery_cachep);
}

/* --------------------------------------------------------------------- */

static void mcgroupmember_add(struct mcgroup *grp,
                              struct mcfw_netdriver *drv,
							  int port,
							  u32 ipaddr)
{
   struct mcgroupmember *p, **pp;
   for (pp = &grp->members; (p = *pp) != 0;  pp = &p->next) {
	  if (p->drv == drv && p->port == port && p->ipaddr == ipaddr)
		 break;
   }

   if ((p = *pp) == 0) {
	  p = (struct mcgroupmember *)kmalloc(sizeof(struct mcgroupmember), GFP_ATOMIC);
	  if (p == 0) {
		 printk(KERN_ERR "mcfw: alloc mcgroupmember failed\n");
		 return;
	  }
	  memset(p, 0, sizeof(struct mcgroupmember));
      p->drv = drv;
      p->port = port;
      p->ipaddr = ipaddr;
	  *pp = p;
	  printk(KERN_INFO "mcgroupmember(%s:%d:%u.%u.%u.%u): %u.%u.%u.%u add\n",
				 drv->name, port, NIPQUAD(grp->group), NIPQUAD(ipaddr));
   } else {
      p->drv = drv;
      p->port = port;
	  printk(KERN_INFO "mcgroupmember(%s:%d:%u.%u.%u.%u): %u.%u.%u.%u refresh\n",
				 drv->name, port, NIPQUAD(grp->group), NIPQUAD(ipaddr));
   }
   p->active = 1;
}

static void mcgroupmember_del(struct mcgroup *grp,
                              struct mcfw_netdriver *drv,
							  int port,
							  u32 ipaddr)
{
   struct mcgroupmember *p, **pp;
   for (pp = &grp->members; (p = *pp) != 0;  pp = &p->next) {
	  if (p->drv == drv && p->port == port && p->ipaddr == ipaddr)
		 break;
   }

   if ((p = *pp) != 0) {
	  *pp = p->next;
	  printk(KERN_INFO "mcgroupmember(%s:%d:%u.%u.%u.%u): %u.%u.%u.%u deleted\n",
				 drv->name, port, NIPQUAD(grp->group), NIPQUAD(ipaddr));
	  kfree(p);
   } else {
	  printk(KERN_INFO "mcgroupmember(%s:%d:%u.%u.%u.%u): %u.%u.%u.%u already deleted\n",
				 drv->name, port, NIPQUAD(grp->group), NIPQUAD(ipaddr));
   }
}

static void mcgroupmember_clear_active(struct mcfw_netdriver *drv,
									   mcfw_portset portset,
                                       struct mcgroup *grp)
{
   struct mcgroupmember *p;
   for (p = grp->members; p; p = p->next)
	  if (p->drv == drv && mcfw_portset_port_is_in_set(&portset, p->port))
	     p->active = 0;
}

static void mcgroupmember_remove_not_active(struct mcgroup *grp)
{
   struct mcgroupmember *p, **pp = &grp->members;
   while ((p = *pp) != 0) {
      if (p->active) {
         pp = &p->next;
      } else {
         *pp = p->next;
         kfree(p);
      }
   }
}

static void mcgroupmember_delete_all(struct mcgroup *grp)
{
   struct mcgroupmember *p, **pp = &grp->members;
   while ((p = *pp) != 0) {
      *pp = p->next;
      kfree(p);
   }
}

/* --------------------------------------------------------------------- */

static inline int grouphashval(u32 group)
{
#ifdef __BIG_ENDIAN
   return ((group & 0xff) ^ ((group >> 8) & 0xff)) % MCGROUP_HASHSIZ;
#else
   return (((group >> 24) & 0xff) ^ ((group >> 16) & 0xff)) % MCGROUP_HASHSIZ;
#endif
}

static inline struct mcgroup **find_mcgroup(u32 group)
{
   struct mcgroup **pp = &grouphashtab[grouphashval(group)];
   for ( ; *pp && (*pp)->group != group; pp = &(*pp)->link)
	  ;
   return pp;
}

static struct mcgroup *find_or_add_mcgroup(u32 group)
{
   struct mcgroup **pp, *p;

   if (LOCAL_MCAST(group))
	  return 0;
   if (LOCALAREA_MCAST(group))
	  return 0;

   pp = find_mcgroup(group);
   if ((p = *pp) == 0) {
      int i;
	  p = (struct mcgroup *)kmalloc(sizeof(struct mcgroup), GFP_ATOMIC);
	  if (p == 0) {
		 printk(KERN_ERR "mcfw: alloc group failed\n");
		 return 0;
	  }
	  memset(p, 0, sizeof(struct mcgroup));
	  p->group = group;
	  p->groupmac[0] = 0x01;
	  p->groupmac[1] = 0x00;
	  p->groupmac[2] = 0x5e;
	  p->groupmac[3] = (u8)(ntohl(group) >> 16) & 0x7f;
	  p->groupmac[4] = (u8)(ntohl(group) >> 8) & 0xff;
	  p->groupmac[5] = (u8)(ntohl(group)) % 0xff;
	  for (i=0; i < MAX_NETDRIVER; i++)
		 p->subscriptions[i].group = group;
	  *pp = p;
   }
   return p;
}

static void mcgroup_check_del(u32 group)
{
   struct mcgroup **pp, *p;

   pp = find_mcgroup(group);
   if ((p = *pp) != 0) {
	  mcgroupmember_remove_not_active(p);
	  if (p->members)
		 return;
	  *pp = p->link;
	  printk(KERN_INFO "mcfw: leave %u.%u.%u.%u\n", NIPQUAD(group));
      kfree(p);
   }
}

static void mcgroup_del(u32 group)
{
   struct mcgroup **pp, *p;

   pp = find_mcgroup(group);
   if ((p = *pp) != 0) {
	  mcgroupmember_delete_all(p);
	  *pp = p->link;
	  printk(KERN_INFO "mcfw: fast leave %u.%u.%u.%u\n", NIPQUAD(group));
      kfree(p);
   }
}
/* --------------------------------------------------------------------- */

static void mcgroup_reconstruct_subscriptions(struct mcgroup *grp)
{
   struct mcgroupmember *p;
   int i;

   for (i=0; i < MAX_NETDRIVER; i++)
	  mcfw_portset_reset(&grp->subscriptions[i].portset);
   for (p = grp->members; p; p = p->next) {
      struct mcsubcription *sub = &grp->subscriptions[p->drv->index];
      mcfw_portset_port_add(&sub->portset, p->port);
   }
}

static void mcfw_add_subscription(struct mcfw_netdriver *drv,
                                  int port, u32 ipaddr, u32 group)
{
   struct mcgroup *p;
   unsigned long flags;

   spin_lock_irqsave(&group_lock, flags);
   if ((p = find_or_add_mcgroup(group)) != 0) {
      struct mcsubcription *sub = &p->subscriptions[drv->index];
      struct mcquery *q;
      mcfw_portset_port_add(&sub->portset, port);
      for (q = sub->queries; q; q = q->next)
         mcfw_portset_port_del(&q->portset, port);
	  mcgroupmember_add(p, drv, port, ipaddr);
   }
   spin_unlock_irqrestore(&group_lock, flags);
}

static void mcfw_del_subscription(struct mcfw_netdriver *drv,
                                  int port, u32 ipaddr, u32 group)
{
   struct mcgroup *p;
   unsigned long flags;

   spin_lock_irqsave(&group_lock, flags);
   if ((p = find_or_add_mcgroup(group)) != 0) {
	  mcgroupmember_del(p, drv, port, ipaddr);
	  if (p->members == 0) {
         mcgroup_del(group);
	  } else {
		 mcgroup_reconstruct_subscriptions(p);
	  }
   }
   spin_unlock_irqrestore(&group_lock, flags);
}

static void mcfw_query_sent(struct mcfw_netdriver *drv,
                            mcfw_portset portset,
							u32 group, unsigned long maxdelay)
{
   struct mcgroup *p;
   unsigned long flags;

#if 1
   {
	  char buf[128];
	  printk(KERN_INFO "mcfw_query_sent: %s:%s:%u.%u.%u.%u %lu\n",
			   drv->name, portset2str(&portset, buf, sizeof(buf)), 
			   NIPQUAD(group), maxdelay);
   }
#endif

   spin_lock_irqsave(&group_lock, flags);
   if (group) {
      if ((p = find_or_add_mcgroup(group)) != 0) {
         struct mcsubcription *sub = &p->subscriptions[drv->index];
		 mcquery_add(sub, portset, maxdelay);
         mcgroupmember_clear_active(drv, portset, p);
	  }
   } else {
	  int i;
	  for (i = 0; i < MCGROUP_HASHSIZ; i++) {
		 for (p = grouphashtab[i]; p; p = p->link) {
             struct mcsubcription *sub = &p->subscriptions[drv->index];
		     mcquery_add(sub, portset, maxdelay);
             mcgroupmember_clear_active(drv, portset, p);
		 }
	  }
   }
   spin_unlock_irqrestore(&group_lock, flags);
}

/* --------------------------------------------------------------------- */

static struct iphdr *find_iph(struct sk_buff *skb)
{
   struct vlan_ethhdr *hdr = (struct vlan_ethhdr *)skb->data;

   if (skb->len < sizeof(struct ethhdr))
	  return 0;

   if (hdr->h_vlan_proto == __constant_htons(ETH_P_8021Q)) {
      if (skb->len < sizeof(struct vlan_ethhdr)+sizeof(struct iphdr))
	     return 0;
	  if (hdr->h_vlan_encapsulated_proto == __constant_htons(ETH_P_IP))
	     return (struct iphdr *)(skb->data+sizeof(struct vlan_ethhdr));
   } else if (hdr->h_vlan_proto == __constant_htons(ETH_P_IP)) {
      if (skb->len < sizeof(struct ethhdr)+sizeof(struct iphdr))
	     return 0;
	  return (struct iphdr *)(skb->data+sizeof(struct ethhdr));
   }
   return 0;
}

void _mcfw_snoop_recv(struct mcfw_netdriver *drv,
                     int port, struct sk_buff *skb)
{
   struct iphdr *iph;
   struct igmphdr *igmph;
   size_t iphlen;
   size_t len;

   if (skb->len < sizeof(struct ethhdr))
	  return;
   if ((iph = find_iph(skb)) == 0)
	  return;
   if (iph->protocol != IPPROTO_IGMP)
	  return;
   if (ip_fast_csum((u8 *)iph, iph->ihl) != 0)
	  return;
   iphlen = (iph->ihl<<2);
   igmph = (struct igmphdr *)(((u8 *)iph)+iphlen);
   len = ((u8 *)iph) - skb->data + iphlen;
   if (skb->len < len + sizeof(struct igmphdr))
	  return;

   switch (igmph->type) {
	  case IGMP_HOST_LEAVE_MESSAGE:
		 mcfw_del_subscription(drv, port, iph->saddr, igmph->group);
		 break;
	  case IGMP_HOST_MEMBERSHIP_REPORT:
		 mcfw_add_subscription(drv, port, iph->saddr, igmph->group);
		 break;
	  case IGMPV2_HOST_MEMBERSHIP_REPORT:
		 mcfw_add_subscription(drv, port, iph->saddr, igmph->group);
		 break;
	  case IGMPV3_HOST_MEMBERSHIP_REPORT:
		 {
            struct igmpv3_report *hp = (struct igmpv3_report *)igmph;
			struct igmpv3_grec *p = &hp->grec[0];
			u16 i;

			len += sizeof(struct igmpv3_report);
            if (skb->len < len) 
			   return;

			for (i = 0; i < ntohs(hp->ngrec); i++) {
               len = ((u8 *)p) - skb->data + sizeof(struct igmpv3_grec);
			   if (skb->len < len) 
				  return;
			   switch (p->grec_type) {
				  case IGMPV3_CHANGE_TO_EXCLUDE:
				  case IGMPV3_MODE_IS_EXCLUDE:
					 mcfw_add_subscription(drv, port, iph->saddr, p->grec_mca);
					 break;
				  case IGMPV3_MODE_IS_INCLUDE:
				  case IGMPV3_CHANGE_TO_INCLUDE:
					 if (p->grec_nsrcs)
		                mcfw_add_subscription(drv, port, iph->saddr, p->grec_mca);
					 else
		                mcfw_del_subscription(drv, port, iph->saddr, p->grec_mca);
					 break;
				  case IGMPV3_ALLOW_NEW_SOURCES:
				  case IGMPV3_BLOCK_OLD_SOURCES:
					 break;
			   }
			   p = (struct igmpv3_grec *)((u8 *)p + sizeof(struct igmpv3_grec));
			   p = (struct igmpv3_grec *)((u8 *)p + p->grec_nsrcs*sizeof(u32));
			}
		 }
		 break;
   }
}

void _mcfw_snoop_send(struct mcfw_netdriver *drv,
                     mcfw_portset portset, struct sk_buff *skb)
{
   struct iphdr *iph;
   struct igmphdr *igmph;
   unsigned long maxdelay; /* in 1/10 sec */
   size_t iphlen;
   u32 group;
   u16 len;

   if (skb->len < sizeof(struct ethhdr))
	  return;
   if ((iph = find_iph(skb)) == 0)
	  return;
   if (iph->protocol != IPPROTO_IGMP)
	  return;
   if (ip_fast_csum((u8 *)iph, iph->ihl) != 0)
	  return;
   iphlen = (iph->ihl<<2);
   igmph = (struct igmphdr *)(((u8 *)iph)+iphlen);
   len = ((u8 *)iph) - skb->data + iphlen;
   if (skb->len < len + sizeof(struct igmphdr))
	  return;
   group = igmph->group;

   if (igmph->type == IGMP_HOST_MEMBERSHIP_QUERY) {
	  len = ((u8 *)igmph - skb->data);
	  if (skb->len == len + sizeof(struct igmphdr)) { /* V1 or V2 */
		 if (igmph->code == 0) { /* V1 */
			maxdelay = 100;
			group = 0;
		 } else { /* V2 */
			maxdelay = igmph->code;
		 }
	  } else if (skb->len < len + sizeof(struct igmpv3_query)) {
		 return; /* illegal packet */
	  } else { /* V3 */
		 struct igmpv3_query *q = (struct igmpv3_query *)igmph;
		 maxdelay = IGMPV3_MRC(q->code);
		 if (!maxdelay) maxdelay = 1;
	  }
	  mcfw_query_sent(drv, portset, group, maxdelay*(HZ/IGMP_TIMER_SCALE));
   }
}

/* --------------------------------------------------------------------- */

int mcfw_multicast_forward_alloc_id(char *name)
{
   int id;
   strncpy((char *)&id, name, sizeof(id));
   return id;
}

void mcfw_multicast_forward_free_id(int id)
{
}

/* --------------------------------------------------------------------- */

struct encapdesc {
   int            enabled;
   u8             mac[ETH_ALEN];
   unsigned short vlanid;        /* hostorder */
   u16            pppoesid;      /* netorder */
};

static struct mcsourcedata {
   atomic_t         index;
   struct encapdesc desc[2];
} mcsourcedata;

void mcfw_multicast_forward_disable(void)
{
   if (atomic_read(&mcsourcedata.index) == 0) {
	  mcsourcedata.desc[1].enabled = 0;
	  atomic_set(&mcsourcedata.index, 0);
   } else {
	  mcsourcedata.desc[0].enabled = 0;
	  atomic_set(&mcsourcedata.index, 1);
   }
   printk(KERN_INFO "mcfw: forward disabled\n");
}

void mcfw_multicast_forward_ethernet_enable(unsigned char mac[ETH_ALEN],
                                            unsigned short vlanid,
											unsigned short pppoesid)
{
   if (atomic_read(&mcsourcedata.index) == 0) {
	  mcsourcedata.desc[1].enabled = 1;
	  memcpy(mcsourcedata.desc[1].mac, mac, ETH_ALEN);
	  mcsourcedata.desc[1].vlanid = vlanid;
	  mcsourcedata.desc[1].pppoesid = htons(pppoesid);
	  atomic_set(&mcsourcedata.index, 1);
   } else {
	  mcsourcedata.desc[0].enabled = 1;
	  memcpy(mcsourcedata.desc[0].mac, mac, ETH_ALEN);
	  mcsourcedata.desc[0].vlanid = vlanid;
	  mcsourcedata.desc[0].pppoesid = htons(pppoesid);
	  atomic_set(&mcsourcedata.index, 0);
   }
   printk(KERN_INFO "mcfw: forward enabled (%02x:%02x:%02x:%02x:%02x:%02x, vlan %hu, pppoe %hu)\n",
			  mac[0], mac[1], mac[2], mac[3], mac[4], mac[5],
			  vlanid, pppoesid);
}

void mcfw_multicast_get_statistic(struct mcfw_statistic *p, int reset)
{
   unsigned long flags;
   spin_lock_irqsave(&group_lock, flags);
   if (p) *p = mcfw_stat;
   if (reset) memset(&mcfw_stat, 0, sizeof(struct mcfw_statistic));
   spin_unlock_irqrestore(&group_lock, flags);
}

static int inline etheraddr_is_multicast(unsigned char mac[ETH_ALEN])
{
   return    mac[0] == 0x01
		  && mac[1] == 0x00
		  && mac[2] == 0x5e
		  && (mac[3] & 0x80) == 0;
}

static inline struct iphdr *
find_iph_for_encap(struct encapdesc *ep, struct sk_buff *skb)
{
   unsigned char *p = skb->data;
   struct pppoe_hdr *pppoeh;
   u16 proto;
   int offset = 0;

   if (!ep->enabled)
	  return 0;

   if (ep->vlanid) {
      struct vlan_ethhdr *vlanh = (struct vlan_ethhdr *)p;
	  if (skb->len < offset+sizeof(struct vlan_ethhdr))
		 return 0;
      if (   memcmp(vlanh->h_dest, ep->mac, ETH_ALEN) != 0
		  && !etheraddr_is_multicast(vlanh->h_dest))
		 return 0;
	  if (vlanh->h_vlan_proto != __constant_htons(ETH_P_8021Q))
		 return 0;
      if ((ntohs(vlanh->h_vlan_TCI) & VLAN_VID_MASK) != ep->vlanid)
		 return 0;
	  p += sizeof(struct vlan_ethhdr);
	  offset += sizeof(struct vlan_ethhdr);
	  proto = vlanh->h_vlan_encapsulated_proto;
   } else {
      struct ethhdr *ethh = (struct ethhdr *)p;
	  if (skb->len < offset+sizeof(struct ethhdr))
		 return 0;
      if (   memcmp(ethh->h_dest, ep->mac, ETH_ALEN) != 0
		  && !etheraddr_is_multicast(ethh->h_dest))
		 return 0;
	  p += sizeof(struct ethhdr);
	  proto = ethh->h_proto;
   }
   if (ep->pppoesid) {
	  if (proto != __constant_htons(ETH_P_PPP_SES))
		 return 0;
	  pppoeh = (struct pppoe_hdr *)p;
	  if (skb->len < offset+sizeof(struct pppoe_hdr)+2+sizeof(struct iphdr))
		 return 0;
	  if (pppoeh->sid != ep->pppoesid)
		 return 0;
	  p += sizeof(struct pppoe_hdr);
	  if (p[0] != 0x00 && p[1] == 0x21)
		 return 0;
	  p += 2;
	  return (struct iphdr *)p;
   } else {
	  if (skb->len < offset+sizeof(struct iphdr))
		 return 0;
	  if (proto != __constant_htons(ETH_P_IP))
		 return 0;
	  return (struct iphdr *)p;
   }
}

int mcfw_multicast_forward_ethernet(int sourceid, struct sk_buff *skb)
{
   mcfw_portset portset[MAX_NETDRIVER];
   struct mcgroup *p;
   struct iphdr *iph;
   unsigned long flags;
   int count = 0;
   int i;

   iph = find_iph_for_encap(
				 &mcsourcedata.desc[atomic_read(&mcsourcedata.index)], skb);

   if (!iph)
	  return 0;

   if (iph->protocol == IPPROTO_UDP && MULTICAST(iph->daddr)) {
      spin_lock_irqsave(&group_lock, flags);
	  if ((p = *find_mcgroup(iph->daddr)) != 0) {
		 for (i = 0; i < MAX_NETDRIVER; i++) {
			if (mcfw_drivers[i]) {
			   portset[i] = p->subscriptions[i].portset;
	           if (!mcfw_portset_empty(&portset[i]))
			      count++;
			} else {
	           mcfw_portset_reset(&portset[i]);
			}
		 }
	  }
	  if (count) {
		 struct ethhdr *ethh;
		 size_t size = ((u8 *)iph) - skb->data - sizeof(struct ethhdr);

		 mcfw_stat.pkt_forwarded++;
		 mcfw_stat.bytes_forwarded += skb->len;

		 if (size) skb_pull(skb, size);
		 ethh = (struct ethhdr *)skb->data;
		 ethh->h_proto = __constant_htons(ETH_P_IP);
		 memcpy(ethh->h_dest, p->groupmac, ETH_ALEN);

         spin_unlock_irqrestore(&group_lock, flags);

		 for (i = 0; i < MAX_NETDRIVER; i++) {
			if (!mcfw_portset_empty(&portset[i])) {
			   if (--count) {
				  struct sk_buff *nskb = skb_copy(skb, GFP_ATOMIC);
				  if (nskb)
					 (*mcfw_drivers[i]->netdriver_mc_transmit)(
											  mcfw_drivers[i]->privatedata,
											  sourceid,
											  portset[i], nskb);
			   } else {
                  (*mcfw_drivers[i]->netdriver_mc_transmit)(
							               mcfw_drivers[i]->privatedata,
								           sourceid,
										   portset[i], skb);
				  break;
			   }
			}
		 }
		 return 1;
	  }
	  mcfw_stat.pkt_dropped++;
	  mcfw_stat.bytes_dropped += skb->len;
	  spin_unlock_irqrestore(&group_lock, flags);
	  dev_kfree_skb_any(skb);
	  return 2;
   }
   return 0;
}

EXPORT_SYMBOL(mcfw_netdriver_register);
EXPORT_SYMBOL(mcfw_netdriver_unregister);
EXPORT_SYMBOL(_mcfw_snoop_recv);
EXPORT_SYMBOL(_mcfw_snoop_send);
EXPORT_SYMBOL(mcfw_multicast_forward_alloc_id);
EXPORT_SYMBOL(mcfw_multicast_forward_free_id);
EXPORT_SYMBOL(mcfw_multicast_forward_disable);
EXPORT_SYMBOL(mcfw_multicast_forward_ethernet_enable);
EXPORT_SYMBOL(mcfw_multicast_get_statistic);
EXPORT_SYMBOL(mcfw_multicast_forward_ethernet);

static int __init mcfw_init_module(void)
{
	int err = mcquery_init();
	if (err) {
	   printk(KERN_ERR "mcfw_init: failed\n");
	} else {
	   printk(KERN_INFO "mcfw_init: ok\n");
	}
	return err;
}

static void __exit mcfw_exit_module(void)
{
   mcquery_exit();
}

module_init(mcfw_init_module);
module_exit(mcfw_exit_module);
